import os
import rosbag
import cv2
import numpy as np
import process_lidar
#import pandas as pd
from cv_bridge import CvBridge
from glob import glob
from bisect import bisect_left
from random import randint

class Value:
    def __init__(self, v=None):
        self.v = v

def make_dirs(folder_path):
    if not os.path.isdir(folder_path):
        os.makedirs(folder_path)

    left_folder = os.path.join(folder_path, 'left')
    right_folder = os.path.join(folder_path, 'right')
    center_folder = os.path.join(folder_path, 'center')
    lidar_folder = os.path.join(folder_path, 'lidar')

    if not os.path.isdir(left_folder):
        os.makedirs(left_folder)
    if not os.path.isdir(right_folder):
        os.makedirs(right_folder)
    if not os.path.isdir(center_folder):
        os.makedirs(center_folder)
    if not os.path.isdir(lidar_folder):
        os.makedirs(lidar_folder)
    return left_folder, right_folder, center_folder, lidar_folder

def closest(time_stamp, t_list):
    pos = bisect_left(t_list, time_stamp)
    if pos == 0:
        return pos
    if pos == len(t_list):
        return pos-1
    before = t_list[pos - 1]
    after = t_list[pos]
    if after - time_stamp < time_stamp - before:
       return pos
    else:
       return pos-1

def get_driving_topic(topics):
    x = [topic for topic in driving_topics if topic in reversed(topics)]
    return x[0] # should only be one topic, but just in case, we take the first element
def get_camera_topics(topics):
    return [topic for topic in camera_topics if topic in reversed(topics)]
def get_lidar_topics(topics):
    x = [topic for topic in lidar_topics if topic in reversed(topics)]
    return x[0] # should only be one topic, but just in case, we take the first element

def is_valid_rosbag(topics):
    camera_center = '/camera/center'
    if any(topic in reversed(topics) for topic in driving_topics):
        if any(topic in reversed(topics) for topic in lidar_topics):
            if camera_center in reversed(topics):
                return True
    return False

def find_valid_rosbags(data_dir): # returns a list with all folders containing valid rosbags
    valid_folders = []
    sub_folders = os.listdir(data_dir)
    for folder in sub_folders:
        data_path = os.path.join(data_dir, folder)
        if not os.path.isdir(data_path):
            continue
        for rosbag_file in glob(os.path.join(data_path, '*.bag')):
            try:
                bag = rosbag.Bag(rosbag_file)
                topics = bag.get_type_and_topic_info()[1].keys()
            except rosbag.bag.ROSBagUnindexedException:
                continue
            except rosbag.bag.ROSBagException:
                continue
            if is_valid_rosbag(topics):
                valid_folders.append(folder)
                break
    return valid_folders

def get_save_folder(topic):
    folder = ""
    if topic == '/camera/center':
        folder = center_folder
    elif topic == '/camera/left' or topic == '/camera/stereo_left':
        folder = left_folder
    elif topic == '/camera/right' or topic == '/camera/stereo_right':
        folder = right_folder
    return folder

def save(path, driving_data, img_id):
    print("... saving operation ...")
    f = open(os.path.join(path, 'data.txt'), 'a+')
    for i in range(len(driving_data)):
        data='img_'+str(img_id[i])+'.png'+'\t'+'lidar_'+str(img_id[i])+'.bin'+'\t'+str(driving_data[i][0])+'\n'
        f.write(data)
    f.close()
    print("... saving operation complete ...")

def get_index(width):
    return randint(0,width-1)

def get_column(index, image):
    if index == 0:
        return image[:,index,:].reshape((OUTPUT_HEIGHT,1,-1))
    elif index == image.shape[1]-1:
        return image[:,image.shape[1]-1,:].reshape((OUTPUT_HEIGHT,1,-1))
    else:
        image_col_1 = image[:,index,:].reshape((OUTPUT_HEIGHT,1,-1))
        image_col_2 = image[:,index+1,:].reshape((OUTPUT_HEIGHT,1,-1))
        col = np.empty(shape=(OUTPUT_HEIGHT, 1, OUTPUT_CHANNELS))
        for i in range(OUTPUT_HEIGHT):
            col[i,0,0] = (image_col_1[i,0,0]+image_col_2[i,0,0])/2.0
            col[i,0,1] = (image_col_1[i,0,1]+image_col_2[i,0,1])/2.0
        return col

def shrink(index, image):
    if index == 0:
        index+=1
    elif index == image.shape[1]-1:
        index-=1
    col = get_column(index, image)
    for i in range(2):
        image = np.delete(image, index, axis=1)
    image = np.insert(image, index, [[col]], axis=1)
    return image

def interpolate(images, output_width):
    scaled_lidar_images = []
    for image in images:
        if image.shape[1] < output_width:
            image_width = image.shape[1]
            for i in range(output_width - image_width):
                index = get_index(image.shape[1])
                col = get_column(index, image)
                image = np.insert(image, index+1, [[col]], axis=1)
            scaled_lidar_images.append(image)
        elif image.shape[1] > output_width:
            image_width = image.shape[1]
            for i in range(image_width - output_width):
                index = get_index(image.shape[1])
                image = shrink(index, image)
            scaled_lidar_images.append(image)
        else:
            scaled_lidar_images.append(image)
    return scaled_lidar_images

def gen_lidar_image(scan):
    try:
        azimuth = [(element.azimuth/100.0) for element in scan]
    except AttributeError:
        azimuth = [(element.angle) for element in scan]

    index_1, index_2 = closest(ANGLE_1, azimuth), closest(ANGLE_2, azimuth)
    scan = scan[index_1:index_2]

    output_width = len(scan)

    lidar_image = np.empty(shape=(OUTPUT_HEIGHT, output_width, OUTPUT_CHANNELS))
    for i in range(output_width):
        for j in range(OUTPUT_HEIGHT):
            try:
                lidar_image[j,i,0] = scan[i].strongest.laser_returns[j].distance
                lidar_image[j,i,1] = scan[i].strongest.laser_returns[j].intensity
            except AttributeError:
                lidar_image[j,i,0] = scan[i].strong.laser_returns[j].distance
                lidar_image[j,i,1] = scan[i].strong.laser_returns[j].intensity
    return lidar_image

def get_lidar_images(lidar_data):
    lidar_images = []
    print("... making lidar images ...")
    for scan in lidar_data:
        lidar_image = gen_lidar_image(scan)
        lidar_images.append(lidar_image)
    print("... done ...")
    return lidar_images

def get_sync_driving_data(driving_data, d_timestamps, c_timestamps):
    sync_driving_data = []
    print("... synchronizing steering angles ...")
    for timestamp in c_timestamps:
        index = closest(timestamp, d_timestamps)
        sync_driving_data.append(driving_data[index])
    print("... done ...")
    return sync_driving_data

def get_sync_lidar_data(lidar_images, l_timestamps, c_timestamps):
    sync_lidar_imgs = []
    print("... synchronizing lidar images ...")
    for timestamp in c_timestamps:
        index = closest(timestamp, l_timestamps)
        sync_lidar_imgs.append(lidar_images[index])
    print("... done ...")

    widths = np.array([image.shape[1] for image in sync_lidar_imgs])
    output_width = np.argmax(np.bincount(widths))
    sync_and_scaled_lidar_images = interpolate(sync_lidar_imgs, output_width)
    return sync_and_scaled_lidar_images

def get_sync_lidar_paths(l_timestamps, c_timestamps):
    sync_paths = []
    for c_timestamp in c_timestamps:
        index = closest(c_timestamp, l_timestamps)
        path = os.path.join(lidar_folder, 'lidar_'+str(index)+'.bin')
        sync_paths.append(path)
    return sync_paths

def get_driving_data(bag, d_topic):
    timestamps, driving_data = [], []
    if d_topic == '/olav/vehicle_status':
        print("... retrieving driving data ...")
        for topic, msg, t in bag.read_messages(topics=[d_topic]):
            timestamps.append(t.secs+t.nsecs*10**(-9))
            driving_data.append((msg.current_steering_angle_deg, msg.speed_ms))
        print("... done ...")
    else:
        # /olav/vehicle_measurements
        print("... retrieving driving data ...")
        for topic, msg, t in bag.read_messages(topics=[d_topic]):
            timestamps.append(t.secs+t.nsecs*10**(-9))
            driving_data.append((msg.drive.steering_angle, msg.drive.speed))
        print("... done ...")
    return timestamps, driving_data

def gen_lidar_images(bag, l_topic):
    timestamps = []
    if l_topic == '/lidar_sweeps':
        print("... making lidar data ...")
        i=0
        for topic, msg, t in bag.read_messages(topics=[l_topic]):
            timestamps.append(t.secs+t.nsecs*10**(-9))
            lidar_image = gen_lidar_image(msg.scans)
            #file_name = 'lidar_'+str(i)+'.npy'
            #np.save(os.path.join(lidar_folder,file_name), lidar_image)
            file_name = 'lidar_'+str(i)+'.bin'
            new_file = open(os.path.join(lidar_folder,file_name), 'wb')
            lidar_image_bytes = bytearray(lidar_image)
            new_file.write(lidar_image_bytes)
            i+=1
        print("... done ...")
    return timestamps

def copy_lidar_images_to(save_folder, sync_lidar_paths, img_id):
    i=0
    print("... synchronizing lidar images ...")
    for path in sync_lidar_paths:
        file_name = 'lidar_'+str(img_id[i])+'.bin'
        lidar_raw = open(path, 'rb').read()

        new_file = open(os.path.join(save_folder, file_name), 'wb')
        new_file.write(lidar_raw)
        i+=1
    print("... done ...")

def clear_lidar_folder():
    files = glob(os.path.join(lidar_folder, '*.bin'))
    for f in files:
        os.remove(f)

def generate(bag, d_topic, c_topics, l_topic):
    global seq

    d_timestamps, driving_data = get_driving_data(bag, d_topic) #steering_angles #s_timestamps
    l_timestamps = gen_lidar_images(bag, l_topic)
    for topic in c_topics:
        print("... topic: " + topic + " ...")
        c_timestamps, img_id = [],[]
        save_folder = get_save_folder(topic)

        i = seq[topic].v
        print("... making dataset for " + topic + " ...")
        for topic, msg, t in bag.read_messages(topics=[topic]):
            c_timestamp = t.secs+t.nsecs*10**(-9)
            index = closest(c_timestamp, d_timestamps)
            speed = float(driving_data[index][1])
            if speed > 1.0:
                name = 'img_'+str(i)+'.png'
                cv_image = CvBridge().imgmsg_to_cv2(msg, desired_encoding="mono8")
                cv2.imwrite(os.path.join(save_folder, name), cv_image)

                c_timestamps.append(c_timestamp)
                img_id.append(i)
                i+=1
        print("... done ...")
        print("... image count: " + str(len(c_timestamps)) + " ...")

        sync_driving_data = get_sync_driving_data(driving_data, d_timestamps, c_timestamps)
        sync_lidar_paths = get_sync_lidar_paths(l_timestamps, c_timestamps)
        copy_lidar_images_to(save_folder, sync_lidar_paths, img_id)

        save(save_folder, sync_driving_data, img_id)

        seq[topic].v = i

def generate_data(data_path):
    for rosbag_file in glob(os.path.join(data_path, '*.bag')):
        try:
            bag = rosbag.Bag(rosbag_file)
            topics = bag.get_type_and_topic_info()[1].keys()
        except rosbag.bag.ROSBagUnindexedException:
            continue
        except rosbag.bag.ROSBagException:
            continue

        if not is_valid_rosbag(topics):
            continue

        print("ROSBAG: " + os.path.basename(rosbag_file))
        d_topic = get_driving_topic(topics)
        c_topics = get_camera_topics(topics)
        l_topic = get_lidar_topics(topics)

        generate(bag, d_topic, c_topics, l_topic)
        clear_lidar_folder()

        bag.close()
        print("CLOSING ROSBAG")
        print("")

ANGLE_1 = 120
ANGLE_2 = 240

OUTPUT_WIDTH = 723
OUTPUT_HEIGHT = 32
OUTPUT_CHANNELS = 2

data_dir = '/home/jodir/olav/sensor_recordings'
training_folder = os.path.join(data_dir, 'data_lidar')

driving_topics = ['/olav/vehicle_status', '/olav/vehicle_measurements']
camera_topics = ['/camera/center','/camera/left','/camera/right','/camera/stereo_left','/camera/stereo_right']
lidar_topics = ['/lidar_sweeps']

#folders = find_valid_rosbags(data_dir)
folders=['2016_10_19', '2016_09_16', '2016_05_09', '2016_09_15', '2016_09_08', '2017_05_10', '2017_02_24', '2017_05_05', '2017_05_09', '2016_04_14', '2016_09_09'] #folders containing valid rosbags
v1, v2, v3 = Value(0), Value(0), Value(0)
seq = {'/camera/center': v1, '/camera/left': v2, '/camera/stereo_left': v2, '/camera/right': v3, '/camera/stereo_right': v3}

left_folder, right_folder, center_folder, lidar_folder = make_dirs(training_folder)

for folder in folders:
    print("#############################")
    print("INSIDE FOLDER: " + folder)
    print("")

    folder_path = os.path.join(data_dir, folder)
    generate_data(folder_path)

    print("")
    print("#############################")

print("#############################")
print("... processing lidar data ...")
for folder in ['center','left', 'right']:
    data_path = os.path.join(data_dir,'data',folder)
    process_lidar.process_lidar(data_path)

    #image_widths = process_lidar.image_widths(data_path)
    #print pd.Series(image_widths).value_counts()
print("... done ...")
print("#############################")
