import os.path
from glob import glob
import re
import numpy as np
import scipy.misc
import eval_segm
from sklearn.metrics import average_precision_score

def elevenPointAP(precision, recall):
    """
    Computes the 11-point interpolated average precision for a list of corresponding precision and recall values.

    Parameters
    ----------
    precision: list
               A list of precision values computed for the retrieved documents against the ground-truth relevant documents
    recall: list
            A list of recall values computed for the retrieved documents against the ground-truth relevant documents

    Returns
    -------
    epap: list
          a list of interpolated average precision at predefined recall values (at [0.0..0.1..1.0])
    """
    epap = []
    start_idx = 0
    for r in np.arange(0,1.1, 0.1):
        for i in range(start_idx, len(precision)):
            if recall[i] == r:
                epap.append(max(precision[i:]))
                start_idx = i+1
                break
            elif (i < len(precision) - 1) and recall[i] < r and recall[i+1] > r:
                epap.append(max(precision[i+1:]))
                start_idx = i+2
                break

    return epap

def precision(actual, predicted):
    """
    Computes the precision values for all sublists of predicted elements starting from the first element.

    Parameters
    ----------
    actual: list
            a list of relevant documents
            (doesn't need to be ordered)
    predicted: list
               an ordered list of predicted documents
               (should be ordered)

    Returns
    -------
    p: list
       a list of precision values for every sublist of predicted documents in predicted[:i]
    """
    p = []

    for i in range(len(predicted)):
        print(i)
        tp = [doc for doc in predicted[:i+1] if doc in actual]
        p.append(len(tp) * 1.0 / (i+1))

    return p

def recall(actual, predicted):
    """
    Computes the recall value of all sublists of predicted elements starting from the first element.

    Parameters
    ----------
    actual: list
            a list of relevant documents
            (doesn't need to be ordered)
    predicted: list
               an ordered list of predicted documents
               (should be ordered)

    Returns
    -------
    r: list
       a list of recal values for every sublist of predicted documents in predicted[:i]
    """
    la = len(actual)
    r = []

    for i in range(len(predicted)):
        tp = [doc for doc in predicted[:i+1] if doc in actual]
        r.append(len(tp) * 1.0 / la)

    return r

data_dir = os.path.abspath('data')
runs_dir = os.path.abspath('runs')

data_folder = os.path.join(data_dir, 'FFI/testing/')


image_paths = glob(os.path.join(runs_dir, '*_binary', '*.png'))
print(os.path.join(data_dir, data_folder, 'labels', 'img_*.png'))

label_paths = {
    os.path.basename(path): path
    for path in glob(os.path.join(data_folder, 'labels', 'img_*.png'))}

image_shape = (256, 320)
pixel_acc, mean_acc, mean_iu, f_w_iu, mean_ap = 0, 0, 0, 0, 0
N = 0

for image_file in image_paths:
    gt_image_file = label_paths[os.path.basename(image_file)]
    image = scipy.misc.imread(image_file)
    gt_image = scipy.misc.imresize(scipy.misc.imread(gt_image_file, mode='L'), image_shape, interp='nearest')

    print(gt_image)
    print(image)

    flat_image = np.array(image).flatten() /255

    flat_gt_image = np.array(gt_image).flatten() / 255

    ap = average_precision_score(flat_gt_image,flat_image)

    print('check0')

    p = precision(flat_gt_image,flat_image)

    print('check1')

    r = recall((flat_gt_image,flat_image))

    print('check2')

    ap11 = elevenPointAP(p,r)

    print('check3')

    p_a = eval_segm.pixel_accuracy(image, gt_image)
    m_a = eval_segm.mean_accuracy(image, gt_image)
    m_iu = eval_segm.mean_IU(image, gt_image)
    f_iu = eval_segm.frequency_weighted_IU(image, gt_image)

    print("Image Name: ", os.path.basename(image_file))
    print("Pixel Accuracy: ", p_a)
    print("Mean Accuracy: ", m_a)
    print("Mean IU: ", m_iu)
    print("Frequency IU: ", f_iu)
    print("Average Precision: ", ap)
    print("11-point AP: ", ap11)
    print("")

    pixel_acc += p_a
    mean_acc += m_a
    mean_iu += m_iu
    f_w_iu += f_iu
    mean_ap += ap


    N+=1

print("-------------------------------------")
print('Average over all {} test images.\n'.format(N))
print("Pixel Accuracy: ", pixel_acc/N)
print("Mean Accuracy: ", mean_acc/N)
print("Mean IU: ", mean_iu/N)
print("Frequency Weighted IU: ", f_w_iu/N)
print("Average Precision: ", mean_ap/N, "\n")
