# Copyright 2016 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Generic evaluation script that evaluates a model using a given dataset."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import math
import tensorflow as tf
import numpy as np
import scipy.io as scio

from datasets import dataset_factory
from nets import nets_factory
from preprocessing import preprocessing_factory

slim = tf.contrib.slim

tf.app.flags.DEFINE_integer(
    'batch_size', 20, 'The number of samples in each batch.')

tf.app.flags.DEFINE_integer(
    'max_num_batches', 1000,
    'Max number of batches to evaluate by default use all.')

tf.app.flags.DEFINE_string(
    'master', '', 'The address of the TensorFlow master to use.')
####==================================================####
tf.app.flags.DEFINE_string(
    'checkpoint_path', '/home/kyb/Deep-Learning-for-UGV-control/multigpu_setup/log/',
    'The directory where the model was written to or an absolute path to a '
    'checkpoint file.')
####==================================================####
tf.app.flags.DEFINE_string(
    'eval_dir', '/home/kyb/Deep-Learning-for-UGV-control/multigpu_setup/eval_log/', 'Directory where the results are saved to.')

tf.app.flags.DEFINE_integer(
    'num_preprocessing_threads', 1,
    'The number of threads used to create the batches.')

tf.app.flags.DEFINE_string(
    'dataset_name', 'ugv_camera', 'The name of the dataset to load.')

tf.app.flags.DEFINE_string(
    'dataset_split_name', 'train', 'The name of the train/test split.')
####==================================================####
tf.app.flags.DEFINE_string(
    'dataset_dir', '/dataset', 'The directory where the dataset files are stored.')

tf.app.flags.DEFINE_integer(
    'labels_offset', 0,
    'An offset for the labels in the dataset. This flag is primarily used to '
    'evaluate the VGG and ResNet architectures which do not use a background '
    'class for the ImageNet dataset.')

tf.app.flags.DEFINE_string(
    'model_name', 'vgg_16', 'The name of the architecture to evaluate.')

tf.app.flags.DEFINE_string(
    'preprocessing_name', None, 'The name of the preprocessing to use. If left '
    'as `None`, then the model_name flag is used.')

tf.app.flags.DEFINE_float(
    'moving_average_decay', None,
    'The decay to use for the moving average.'
    'If left as None, then moving averages are not used.')

tf.app.flags.DEFINE_integer(
    'eval_image_size', None, 'Eval image size')

tf.app.flags.DEFINE_boolean(
    'gen_features', True, 'Generate features and writes to a mat file')
tf.app.flags.DEFINE_boolean(
    'gen_stats', True, 'Generate mean and squared mean of the feature vectors')


FLAGS = tf.app.flags.FLAGS


def main(_):
  if not FLAGS.dataset_dir:
    raise ValueError('You must supply the dataset directory with --dataset_dir')

  tf.logging.set_verbosity(tf.logging.INFO)
  with tf.Graph().as_default():
    tf_global_step = slim.get_or_create_global_step()

    ######################
    # Select the dataset #
    ######################
    dataset = dataset_factory.get_dataset(
        FLAGS.dataset_name, FLAGS.dataset_split_name, FLAGS.dataset_dir)



    ####################
    # Select the model #
    ####################
    network_fn = nets_factory.get_network_fn(
        FLAGS.model_name,
        num_classes=(dataset.num_classes - FLAGS.labels_offset),
        is_training=False)

    ##############################################################
    # Create a dataset provider that loads data from the dataset #
    ##############################################################
    provider = slim.dataset_data_provider.DatasetDataProvider(
        dataset,
        shuffle=False,
        common_queue_capacity=2 * FLAGS.batch_size,
        common_queue_min=FLAGS.batch_size)
    [image, label, target,index] = provider.get(['image', 'label','target', 'index'])
    label -= FLAGS.labels_offset


    #####################################
    # Select the preprocessing function #
    #####################################
    preprocessing_name = FLAGS.preprocessing_name or FLAGS.model_name
    image_preprocessing_fn = preprocessing_factory.get_preprocessing(
        preprocessing_name,
        is_training=False)

    eval_image_size = FLAGS.eval_image_size or network_fn.default_image_size

    image = image_preprocessing_fn(image, eval_image_size, eval_image_size)

    images, labels,targets,indexes = tf.train.batch(
        [image, label,target,index],
        batch_size=FLAGS.batch_size,
        num_threads=FLAGS.num_preprocessing_threads,
        capacity=5 * FLAGS.batch_size)

    ####################
    # Define the model #
    ####################
    logits,end_points,pool4, pool5 = network_fn(images)

    net = tf.contrib.layers.conv2d(pool5, 256, 3, stride=1, padding='SAME', activation_fn=tf.nn.relu, scope='ntnu1')
    net = tf.contrib.layers.batch_norm(net, is_training=False, scope='ntnu2')
    net = tf.contrib.layers.conv2d(net, 128, 1, stride=1, padding='SAME', activation_fn=tf.nn.relu,scope='ntnu3')
    net = tf.contrib.layers.batch_norm(net, is_training=False, scope='ntnu4')
    net = tf.contrib.layers.conv2d(net, 256, 3, stride=1, padding='SAME', activation_fn=tf.nn.relu,scope='ntnu5')
    net = tf.contrib.layers.batch_norm(net, is_training=False, scope='ntnu6')
    net = tf.contrib.layers.max_pool2d(net, 2, stride=2, padding='SAME', scope='ntnu7')

    net = tf.contrib.layers.conv2d(net, 128, 3, stride=1, padding='SAME', activation_fn=tf.nn.relu,scope='ntnu8')
    net = tf.contrib.layers.batch_norm(net, is_training=False, scope='ntnu9')
    net = tf.contrib.layers.max_pool2d(net, 2, stride=2, scope='ntnu10')

    net = tf.contrib.layers.conv2d(net, 256, 3, stride=1, padding='SAME', activation_fn=tf.nn.relu, scope='ntnu11')
    net = tf.contrib.layers.batch_norm(net, is_training=False, scope='ntnu12')
    net = tf.contrib.layers.conv2d(net, 512, 4, stride=1, padding='SAME', activation_fn=tf.nn.relu, scope='ntnu13')
    net = tf.contrib.layers.batch_norm(net, scope='ntnu14')

    net = tf.contrib.layers.flatten(net)
    final_target = tf.contrib.layers.fully_connected(net, 1, activation_fn=None, scope='ntnu15')

    if FLAGS.moving_average_decay:
      variable_averages = tf.train.ExponentialMovingAverage(
          FLAGS.moving_average_decay, tf_global_step)
      variables_to_restore = variable_averages.variables_to_restore(
          slim.get_model_variables())
      variables_to_restore[tf_global_step.op.name] = tf_global_step
    else:
      variables_to_restore = slim.get_variables_to_restore()




    labels = tf.squeeze(labels)
    targets = tf.squeeze(targets)
    indexes = tf.squeeze(indexes)
    final_target=tf.squeeze(final_target)

    pred_error=targets-final_target

    target_mean= tf.reduce_mean(targets,0)
    error_sqmean=tf.reduce_mean(tf.multiply(pred_error, pred_error),0)
    #print('####################### ',features.get_shape())


    # TODO(sguada) use num_epochs=1
    if FLAGS.max_num_batches:
      num_batches = FLAGS.max_num_batches
    else:
      # This ensures that we make a single pass over all of the data.
      num_batches = math.ceil(dataset.num_samples / float(FLAGS.batch_size))

    if tf.gfile.IsDirectory(FLAGS.checkpoint_path):
      checkpoint_path = tf.train.latest_checkpoint(FLAGS.checkpoint_path)
    else:
      checkpoint_path = FLAGS.checkpoint_path

    variables_to_restore = slim.get_variables_to_restore()
    saver = tf.train.Saver(variables_to_restore)

    def restore_fn(sess):
        return saver.restore(sess, checkpoint_path)

    # Get your supervisor
    sv = tf.train.Supervisor(logdir=FLAGS.eval_dir, summary_op=None, saver=None, init_fn=restore_fn)


    targets_concat =[]
    labels_concat=[]
    indexes_concat=[]
    final_target_concat=[]
    '''probabilities_concat=np.empty((0,dataset.num_classes),float)
    features_concat=np.empty((0,fdim1),float)
    features_mean_concat=np.zeros((1,fdim1),float)
    features_sqmean_concat=np.zeros((1,fdim1),float)'''
    features_divisor=0
    target_mean_concat=0
    error_sqmean_concat=0


    # Now we are ready to run in one session
    with sv.managed_session() as sess:
        for step in range(num_batches):
            sess.run(sv.global_step)


            if (step % 10 == 0):
                print('step %d of Total %d' %(step, num_batches))
            if (FLAGS.gen_features):
                targets_val, labels_val,indexes_val, final_target_val = sess.run([targets, labels,indexes, final_target])
                targets_concat=np.append(targets_concat, targets_val, axis=0)
                labels_concat=np.append(labels_concat, labels_val, axis=0)
                indexes_concat = np.append(indexes_concat, indexes_val, axis=0)
                final_target_concat = np.append(final_target_concat, final_target_val, axis=0)


            if (FLAGS.gen_stats):
                target_mean_val, error_sqmean_val, labels_val=sess.run([target_mean, error_sqmean,labels])
                idim0=labels_val.shape[0]
                if (idim0 == FLAGS.batch_size):
                    target_mean_concat=target_mean_concat+target_mean_val
                    error_sqmean_concat= error_sqmean_concat+error_sqmean_val
                    features_divisor=features_divisor+1

        if (FLAGS.gen_features):
            scio.savemat(FLAGS.eval_dir+'/results.mat', mdict={'targets': targets_concat, 'labs': labels_concat,'indexes': indexes_concat, 'pred_targets': final_target_concat})

        if (FLAGS.gen_stats):
            target_mean_concat=np.divide(target_mean_concat,features_divisor)
            error_sqmean_concat=np.divide(error_sqmean_concat,features_divisor)
            print(target_mean_concat)
            print(error_sqmean_concat)





if __name__ == '__main__':
  tf.app.run()
